package com.sencorsta.ids.core.processor;

import cn.hutool.core.lang.UUID;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.json.JSON;
import cn.hutool.json.JSONUtil;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.sencorsta.ids.core.application.master.response.JoinMasterResponse;
import com.sencorsta.ids.core.application.proxy.ProxyClient;
import com.sencorsta.ids.core.config.ConfigGroup;
import com.sencorsta.ids.core.config.GlobalConfig;
import com.sencorsta.ids.core.constant.ErrorCodeConstant;
import com.sencorsta.ids.core.constant.ProtocolTypeConstant;
import com.sencorsta.ids.core.constant.SerializeTypeConstant;
import com.sencorsta.ids.core.entity.Client;
import com.sencorsta.ids.core.entity.ErrorCode;
import com.sencorsta.ids.core.entity.HttpFileResponse;
import com.sencorsta.ids.core.entity.IdsRequest;
import com.sencorsta.ids.core.entity.IdsResponse;
import com.sencorsta.ids.core.entity.MethodProxy;
import com.sencorsta.ids.core.net.handle.ContentTypes;
import com.sencorsta.ids.core.net.handle.HttpContextHelper;
import com.sencorsta.ids.core.net.protocol.HttpMessage;
import com.sencorsta.ids.core.net.protocol.RpcMessage;
import com.sencorsta.utils.object.Jsons;
import com.sencorsta.utils.string.StringUtil;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelProgressiveFuture;
import io.netty.channel.ChannelProgressiveFutureListener;
import io.netty.handler.codec.http.DefaultFullHttpResponse;
import io.netty.handler.codec.http.DefaultHttpResponse;
import io.netty.handler.codec.http.FullHttpRequest;
import io.netty.handler.codec.http.FullHttpResponse;
import io.netty.handler.codec.http.HttpHeaderNames;
import io.netty.handler.codec.http.HttpHeaderValues;
import io.netty.handler.codec.http.HttpMethod;
import io.netty.handler.codec.http.HttpResponse;
import io.netty.handler.codec.http.HttpResponseStatus;
import io.netty.handler.codec.http.HttpUtil;
import io.netty.handler.codec.http.HttpVersion;
import io.netty.handler.codec.http.LastHttpContent;
import io.netty.handler.codec.rtsp.RtspHeaderValues;
import io.netty.handler.stream.ChunkedFile;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;

import javax.activation.MimetypesFileTypeMap;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.InetSocketAddress;
import java.net.URI;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static io.netty.handler.codec.http.HttpHeaderNames.CONNECTION;
import static io.netty.handler.codec.http.HttpHeaderNames.CONTENT_TYPE;
import static io.netty.handler.codec.http.HttpResponseStatus.OK;
import static io.netty.handler.codec.http.HttpVersion.HTTP_1_1;

/**
 * 消息分发器
 *
 * @author ICe
 */
@Slf4j
@AllArgsConstructor
public class MessageHttpDispatcher implements Runnable {
    private final HttpMessage message;

    @Override
    public void run() {
        log.info("开始处理HTTP消息: {}", message.toStringPlus());
        if (GlobalConfig.IS_DEBUG) {
            long sTime = System.currentTimeMillis();
            String method = message.getMethod();
            ScheduledFuture<?> schedule = MessageProcessor.MONITOR.schedule(() -> {
                log.warn("消息处理超时:{}", method);
            }, 10, TimeUnit.SECONDS);
            execute(message);
            schedule.cancel(true);
            if (System.currentTimeMillis() - sTime > 100) {
                log.warn("消息处理时间过高{} -> {}", method, System.currentTimeMillis() - sTime);
            }
        } else {
            execute(message);
        }
    }

    private void execute(HttpMessage message) {
        HttpMethod method = message.getRequest().method();
        switch (method.name()) {
            // 服务器之间的请求 理论上必须能找到处理器 找不到就返回错误
            case "POST":
            case "GET":
                handleHttp(message);
                break;
            // 推送一般直接触发对应的逻辑就行了
            case "OPTIONS":
                handleHttpOptions(message);
                break;
            default:
                log.warn("不支持协议类型：{}", method);
                break;
        }
//        log.trace("HttpMessage 连接关闭");
//        message.getChannel().close();
    }

    private void handleHttpOptions(HttpMessage message) {
        FullHttpResponse response = new DefaultFullHttpResponse(HTTP_1_1, OK);
        response.headers().set("Access-Control-Allow-Origin", "*");
        response.headers().set("Access-Control-Allow-Credentials", "true");
        response.headers().set("Access-Control-Allow-Methods", "*");
        response.headers().set("Access-Control-Allow-Headers", "Content-Type,Access-Token");
        response.headers().set("Access-Control-Expose-Headers", "*");
        Channel ctx = message.getChannel();
        ctx.write(response);
        ctx.flush();
        ctx.close();
    }

    private void handleHttp(HttpMessage message) {
        //先尝试本地是否有此方法
        boolean over = handleHttpReq(message);
        if (!over) {
            try {
                //本地没有就尝试代理给远程服务器
                String method = message.getMethod();
                String[] urlPath = extractParts(method);
                log.debug("urlPath:{}", urlPath);
                String desServer = urlPath[0];
                if (ProxyClient.hasServer(desServer)) {
                    RpcMessage rpcMessage = new RpcMessage(ProtocolTypeConstant.TYPE_PROXY_HTTP_REQ);
                    Channel channel = message.getChannel();
                    Client client = channel.attr(GlobalConfig.CLIENT_KEY).get();
                    if (client == null) {
                        //用户没有登陆 生成临时用户id
                        String userId = "TEMP_" + UUID.fastUUID();
                        rpcMessage.setUserId(userId);
                        Client newClient = new Client();
                        newClient.setUserId(userId);
                        newClient.setChannel(channel);
                        GlobalContainer.CLIENTS.put(userId, newClient);
                        channel.attr(GlobalConfig.CLIENT_KEY).set(newClient);
                    } else {
                        rpcMessage.setUserId(client.getUserId());
                    }
                    rpcMessage.setMethod(urlPath[1]);
                    rpcMessage.setSerializeType(SerializeTypeConstant.TYPE_JSON);
                    ObjectNode data = HttpContextHelper.parseParam(message.getRequest());
                    rpcMessage.setData(Jsons.serializeBytes(data));
                    RpcMessage resSync = ProxyClient.requestByTypeSync(rpcMessage, desServer);
                    if (resSync != null) {
                        //如果是文件类型就改成文件响应
                        IdsResponse<HttpFileResponse> idsResponse = Jsons.toBean(resSync.getData(), new TypeReference<IdsResponse<HttpFileResponse>>() {
                        });
                        if (idsResponse != null && idsResponse.getData() != null && ContentTypes.APPLICATION_OCTET_STREAM.equals(idsResponse.getData().getContentType())) {
                            HttpFileResponse httpFileResponse = idsResponse.getData();
                            String contentType = httpFileResponse.getContentType();
                            String filename = httpFileResponse.getFileName();
                            String filePath = httpFileResponse.getFilePath();
                            responseFile(message, filePath, contentType, filename);
                        } else {
                            String contentType = ContentTypes.JSON;
                            byte[] resSyncData = resSync.getData();
                            response(message, resSyncData, contentType);
                        }
                    }
                    over = true;
                }
            } catch (IllegalArgumentException e) {
                log.debug("没有找到对应的代理方法:{}", message.getMethod());
            } catch (Exception e) {
                log.error("本地没有就尝试代理给远程服务器:", e);
                return;
            }
        }
        if (!over) {
            //没有处理器尝试检查本地文件
            if (GlobalConfig.instance().getBool("httpServer.openStaticWeb", ConfigGroup.server.getName(), false)) {
                over = handleStaticFile(message);
            }
        }
        if (!over) {
            Channel ctx = message.getChannel();
            FullHttpRequest request = message.getRequest();
            InetSocketAddress inetSocket = (InetSocketAddress) ctx.remoteAddress();
            String clientIP = inetSocket.getAddress().getHostAddress();
            log.warn("没有找到处理句柄:{}  -> {} IP:{}", request.method(), request.uri(), clientIP);
        }
    }

    private static String[] extractParts(String url) {
        // 定义正则表达式
        String regex = "^/(\\w+)/(.+)$";
        Pattern pattern = Pattern.compile(regex);

        // 创建 Matcher 对象
        Matcher matcher = pattern.matcher(url);

        // 判断是否匹配
        if (matcher.matches()) {
            // 提取匹配的组
            String service = matcher.group(1);
            String path = "/" + matcher.group(2);
            return new String[]{service, path};
        } else {
            // 如果不匹配，可能是输入的 URL 格式不正确，可以根据需要处理异常情况
            throw new IllegalArgumentException("Invalid URL format");
        }
    }

    private boolean handleStaticFile(HttpMessage message) {
        try {
            FullHttpRequest request = message.getRequest();
            ChannelHandlerContext ctx = message.getChannelHandlerContext();
            String path = message.getMethod();
            URI uri = new URI(request.uri());
            log.trace("没有handler 尝试找本地文件 url：{}", uri);
            // 根据路径地址构建文件
            String basePath = System.getProperty("user.dir") + GlobalConfig.instance().getStr("httpServer.pathStaticWeb", ConfigGroup.server.getName(), "/web");
            String pathFile = basePath + URLDecoder.decode(uri.getPath(), "UTF-8");

            File html = new File(pathFile);

            // 状态为1xx的话，继续请求
            if (HttpUtil.is100ContinueExpected(request)) {
                FullHttpResponse response = new DefaultFullHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.CONTINUE);
                ctx.writeAndFlush(response);
            }
            //如果是目录就加默认index.html
            if (html.isDirectory()) {
                html = new File(pathFile + "/index.html");
            }
            // 当文件不存在的时候，将资源指向NOT_FOUND
            boolean isNotFound = false;
            if (!html.exists()) {
                // 构建404页面
                String path404 = basePath + "/404.html";
                html = new File(path404);
                if (!html.exists()) {
                    InetSocketAddress inSocket = (InetSocketAddress) ctx.channel().remoteAddress();
                    String clientIp = inSocket.getAddress().getHostAddress();
                    log.warn("没有找到页面:{} : {} IP:{}", path404, request.uri(), clientIp);
                    InputStream resourceStream = getClass().getClassLoader().getResourceAsStream("fileNotFind.html");
                    ByteBuf content = ctx.alloc().buffer();
                    try {
                        byte[] buffer = new byte[1024];
                        int bytesRead;
                        while ((bytesRead = resourceStream.read(buffer)) != -1) {
                            content.writeBytes(buffer, 0, bytesRead);
                        }
                    } catch (IOException e) {
                        return false;
                    } finally {
                        resourceStream.close();
                    }
                    FullHttpResponse response = new DefaultFullHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.OK, content);
                    response.headers().set(HttpHeaderNames.CONTENT_TYPE, "text/html; charset=UTF-8");
                    ctx.writeAndFlush(response).addListener(ChannelFutureListener.CLOSE);
                    return false;
                } else {
                    isNotFound = true;
                }
            }

            RandomAccessFile file = new RandomAccessFile(html, "r");
            HttpResponse response = new DefaultHttpResponse(HTTP_1_1, HttpResponseStatus.OK);

            // 文件没有发现设置状态为404
            if (isNotFound) {
                response.setStatus(HttpResponseStatus.NOT_FOUND);
            }
            log.trace("找到本地文件 url:{}", uri);
            // 设置文件格式内容
            if (path.endsWith(".html")) {
                response.headers().set("Content-Type", "text/html; charset=UTF-8");
            } else if (path.endsWith(".js")) {
                response.headers().set("Content-Type", "application/x-javascript");
            } else if (path.endsWith(".css")) {
                response.headers().set("Content-Type", "text/css; charset=UTF-8");
            }
            boolean keepAlive = HttpUtil.isKeepAlive(request);
            if (keepAlive) {
                response.headers().set("Content-Length", file.length());
                response.headers().set("Connection", "keep-alive");
            }
            ctx.write(response);
            //通过Netty的ChunkedFile对象直接将文件写入到发送缓冲区中
            ChannelFuture sendFileFuture = ctx.writeAndFlush(new ChunkedFile(file, 0, file.length(), 8192), ctx.newProgressivePromise());
            //为sendFileFuture添加监听器，如果发送完成打印发送完成的日志
            sendFileFuture.addListener(new ChannelProgressiveFutureListener() {
                @Override
                public void operationProgressed(ChannelProgressiveFuture future, long progress, long total) {
                    if (total < 0) {
                        log.trace(uri + " 传输中: " + progress);
                    } else {
                        log.trace(uri + " 传输中: " + progress + "/" + total);
                    }
                }

                @Override
                public void operationComplete(ChannelProgressiveFuture future)
                        throws Exception {
                    log.debug(uri + " 传输完成！");
                    file.close();
                }
            });
            //如果使用chunked编码，最后需要发送一个编码结束的空消息体，将LastHttpContent.EMPTY_LAST_CONTENT发送到缓冲区中，
            //来标示所有的消息体已经发送完成，同时调用flush方法将发送缓冲区中的消息刷新到SocketChannel中发送
            log.trace("开始发送最后一帧信息 ctx.writeAndFlush(LastHttpContent.EMPTY_LAST_CONTENT)");
            ChannelFuture lastContentFuture = ctx.writeAndFlush(LastHttpContent.EMPTY_LAST_CONTENT);
            //如果是非keepAlive的，最后一包消息发送完成后，服务端要主动断开连接
            if (!keepAlive) {
                log.trace("非keepAlive断开连接 lastContentFuture.addListener(ChannelFutureListener.CLOSE)");
                lastContentFuture.addListener(ChannelFutureListener.CLOSE);
            }
            return true;
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }
        return false;
    }

    private boolean handleHttpReq(HttpMessage message) {
        MethodProxy methodProxy = MessageProcessor.getMETHOD_MAP().get(message.getMethod());
        if (ObjectUtil.isNotNull(methodProxy)) {
            try {
                Object result = invoke(message, methodProxy);
                String contentType = ContentTypes.JSON;
                boolean isFile = false;
                if (result instanceof IdsResponse) {
                    IdsResponse<?> idsResponse = (IdsResponse<?>) result;
                    if (idsResponse.getData() instanceof HttpFileResponse) {
                        HttpFileResponse response = ((HttpFileResponse) idsResponse.getData());
                        contentType = response.getContentType();
                        String filename = response.getFileName();
                        String filePath = response.getFilePath();
                        isFile = true;
                        responseFile(message, filePath, contentType, filename);
                    }
                }
                if (!isFile) {
                    byte[] bytes = Jsons.mapper.writeValueAsBytes(result);
                    response(message, bytes, contentType);
                }
                return true;
            } catch (Exception e) {
                log.error(e.getMessage(), e);
            }
        } else {
            log.debug("没有找到对应的处理方法:{}", message.getMethod());
        }
        return false;
    }

    private void responseFile(HttpMessage message, String filePath, String contentType, String fileName) throws Exception {
        FullHttpRequest req = message.getRequest();
        Channel ctx = message.getChannel();
        //文件下传
        File file = new File(filePath);
        RandomAccessFile randomAccessFile = new RandomAccessFile(file, "r");
        long fileLength = randomAccessFile.length();
        log.debug("file:{}", file.getAbsolutePath());
        log.debug("fileLength:{}", fileLength);

        HttpResponse response = new DefaultHttpResponse(HTTP_1_1, OK);
        response.headers().set("Content-Length", fileLength);

        if (contentType.equals(ContentTypes.AUTO)) {
            MimetypesFileTypeMap mimetypesTypeMap = new MimetypesFileTypeMap();
            response.headers().set(CONTENT_TYPE, mimetypesTypeMap.getContentType(file.getPath()));
        } else {
            response.headers().set("Content-Type", contentType);
            String userAgent = req.headers().get("User-Agent");
            String encoderFilename;
            if (userAgent.toUpperCase().contains("MSIE") || userAgent.contains("Trident/7.0")) {
                encoderFilename = URLEncoder.encode(fileName, "UTF-8");
            } else if (userAgent.toUpperCase().contains("MOZILLA") || userAgent.toUpperCase().contains("CHROME")) {
                encoderFilename = new String(fileName.getBytes(), StandardCharsets.ISO_8859_1);
            } else {
                encoderFilename = URLEncoder.encode(fileName, "UTF-8");
            }
            log.debug("encoderFilename:{}", encoderFilename);
            response.headers().set("content-disposition", "attachment;filename=" + encoderFilename);
        }

        response.headers().set("Access-Control-Allow-Origin", "*");
        response.headers().set("Charset", "UTF-8");
        if (HttpUtil.isKeepAlive(req)) {
            response.headers().set(CONNECTION, HttpHeaderValues.KEEP_ALIVE);
        }
        ctx.write(response);
        ChannelFuture sendFileFuture;
        //通过Netty的ChunkedFile对象直接将文件写入到发送缓冲区中
        sendFileFuture = ctx.write(new ChunkedFile(randomAccessFile, 0, fileLength, 8192), ctx.newProgressivePromise());
        //为sendFileFuture添加监听器，如果发送完成打印发送完成的日志
        sendFileFuture.addListener(new ChannelProgressiveFutureListener() {
            @Override
            public void operationProgressed(ChannelProgressiveFuture future, long progress, long total) {
                if (total < 0) {
                    log.trace(fileName + " 传输中: " + progress);
                } else {
                    log.trace(fileName + " 传输中: " + progress + "/" + total);
                }
            }

            @Override
            public void operationComplete(ChannelProgressiveFuture future) {
                log.debug(fileName + " 传输完成！");
            }
        });
        //如果使用chunked编码，最后需要发送一个编码结束的空消息体，将LastHttpContent.EMPTY_LAST_CONTENT发送到缓冲区中，
        //来标示所有的消息体已经发送完成，同时调用flush方法将发送缓冲区中的消息刷新到SocketChannel中发送
        ChannelFuture lastContentFuture = ctx.
                writeAndFlush(LastHttpContent.EMPTY_LAST_CONTENT);
        //如果是非keepAlive的，最后一包消息发送完成后，服务端要主动断开连接
        if (!HttpUtil.isKeepAlive(req)) {
            lastContentFuture.addListener(ChannelFutureListener.CLOSE);
        }
    }

    private void response(HttpMessage message, byte[] bytes, String contentType) throws ErrorCode {
        if (bytes == null) {
            throw ErrorCodeConstant.SYSTEM_ERROR;
        }
        FullHttpResponse response = new DefaultFullHttpResponse(HTTP_1_1, HttpResponseStatus.OK, Unpooled.wrappedBuffer(bytes));
        if (StringUtil.isEmpty(contentType)) {
            response.headers().set("Content-Type", ContentTypes.TEXT_PLAIN);
        } else {
            response.headers().set("Content-Type", contentType);
        }
        response.headers().set("Content-Length", response.content().readableBytes());
        response.headers().set("Access-Control-Allow-Origin", "*");
        if (HttpUtil.isKeepAlive(message.getRequest())) {
            response.headers().set("Connection", RtspHeaderValues.KEEP_ALIVE);
        }
        message.getChannel().write(response);
        message.getChannel().flush();
    }

    private Object invoke(HttpMessage message, MethodProxy methodProxy) throws Exception {
        final Method method = methodProxy.getMethod();
        Class<?> valueType = methodProxy.getValueType();

        Object object;
        if (valueType.isArray()) {
            object = message.getRequest().content().array();
        } else if (valueType == String.class) {
            object = new String(message.getRequest().content().array());
        } else {
            ObjectNode data = HttpContextHelper.parseParam(message.getRequest());
            object = Jsons.getMapper().convertValue(data, valueType);
        }
        if (object == null) {
            object = new Object();
        }
        IdsRequest<?> idsRequest = new IdsRequest<>(object);
        idsRequest.setChannel(message.getChannel());
        idsRequest.setUserId(message.getUserId());
        try {
            return method.invoke(methodProxy.getObj(), idsRequest);
        } catch (Exception exception) {
            if (exception instanceof InvocationTargetException) {
                Throwable targetException = ((InvocationTargetException) exception).getTargetException();
                if (targetException instanceof ErrorCode) {
                    return new IdsResponse<>(null, (ErrorCode) targetException);
                } else {
                    throw exception;
                }
            } else {
                throw exception;
            }
        }
    }
}
