package com.sencorsta.idsoos.service.impl;

import cn.hutool.core.io.FileUtil;
import cn.hutool.core.io.IoUtil;
import cn.hutool.core.lang.UUID;
import cn.hutool.core.util.StrUtil;
import cn.hutool.crypto.digest.DigestUtil;
import cn.hutool.json.JSONObject;
import cn.hutool.jwt.JWT;
import cn.hutool.jwt.JWTUtil;
import com.sencorsta.ids.ErrorCodeList;
import com.sencorsta.ids.api.request.ReqOosMyFiles;
import com.sencorsta.ids.api.request.ReqOosUploadFile;
import com.sencorsta.ids.api.request.ReqUserVerify;
import com.sencorsta.ids.api.response.ResOosMyFiles;
import com.sencorsta.ids.api.response.ResOosUploadFile;
import com.sencorsta.ids.api.response.ResUserVerify;
import com.sencorsta.ids.api.tools.UserReqTools;
import com.sencorsta.ids.bean.obj.FileInfo;
import com.sencorsta.ids.bean.table.FileDictionary;
import com.sencorsta.ids.core.config.ConfigGroup;
import com.sencorsta.ids.core.config.GlobalConfig;
import com.sencorsta.ids.core.entity.ErrorCode;
import com.sencorsta.ids.core.entity.FileUpload;
import com.sencorsta.ids.core.entity.IdsRequest;
import com.sencorsta.ids.core.entity.IdsResponse;
import com.sencorsta.ids.core.entity.annotation.Autowired;
import com.sencorsta.ids.core.entity.annotation.Service;
import com.sencorsta.idsoos.dao.IdsOosDao;
import com.sencorsta.idsoos.service.IdsOosService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;

import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;


/**
 * @author ICe
 */
@Service
@Slf4j
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class IdsOosServiceImpl implements IdsOosService {

    private final IdsOosDao idsOosDao;

    @Override
    public IdsResponse<ResOosUploadFile> uploadFile(IdsRequest<ReqOosUploadFile> request) throws Exception {
        ReqOosUploadFile reqOosUploadFile = request.getData();
        checkToken(reqOosUploadFile.getToken());
        String userId = JWTUtil.parseToken(reqOosUploadFile.getToken()).getPayloads().getStr("userId");

        String path = reqOosUploadFile.getPath();
        if (StrUtil.isBlank(path)) {
            path = "/";
        }
        boolean isPublic = "on".equals(reqOosUploadFile.getIsPublic());

        ResOosUploadFile result = new ResOosUploadFile();
        List<FileInfo> fileInfos = new ArrayList<>();
        result.setFileInfos(fileInfos);

        List<FileUpload> files = reqOosUploadFile.getFiles();
        try {
            for (FileUpload file : files) {
                String fileName = file.getName();
                byte[] fileData = file.getData();

                String uuid = UUID.fastUUID().toString();
                // 计算文件的 MD5 码
                String md5 = DigestUtil.md5Hex(fileData);
                // 保存文件到本地
                String physicalAddress = saveFileToLocal(md5, fileData);

                FileDictionary fileDictionary = new FileDictionary();
                fileDictionary.setUserId(userId);
                fileDictionary.setUuid(uuid);
                fileDictionary.setPath(path);
                fileDictionary.setPhysicalAddress(physicalAddress);
                fileDictionary.setMd5(md5);
                fileDictionary.setFileName(fileName);
                fileDictionary.setPublic(isPublic);

                idsOosDao.insertNewFile(fileDictionary);

                FileInfo fileInfo = new FileInfo();
                fileInfo.setUuid(uuid);
                String host = GlobalConfig.instance().getStr("host.public", ConfigGroup.server.getName(), "0.0.0.0");
                String port = GlobalConfig.instance().getStr("port", ConfigGroup.server.getName(), "80");
                fileInfo.setUrl(StrUtil.format("http://{}:{}/oos/{}", host, port, md5));
                fileInfo.setFileName(path + "/" + fileName);
                fileInfo.setPublic(isPublic);

                // 将文件名和对应的 MD5 码加入结果中
                fileInfos.add(fileInfo);
            }
        } catch (Exception e) {
            log.error("文件保存失败:", e);
            throw ErrorCodeList.FILE_SAVE_FAILED;
        }

        return new IdsResponse<>(result);
    }

    @Override
    public IdsResponse<ResOosMyFiles> myFiles(IdsRequest<ReqOosMyFiles> request) throws Exception {
        ReqOosMyFiles reqOosMyFiles = request.getData();
        checkToken(reqOosMyFiles.getToken());
        String userId = JWTUtil.parseToken(reqOosMyFiles.getToken()).getPayloads().getStr("userId");

        String fileName = reqOosMyFiles.getFileName();
        String uuid = reqOosMyFiles.getUuid();
        String path = reqOosMyFiles.getPath();

        List<FileDictionary> files = idsOosDao.findFiles(userId, fileName, uuid, path);
        String host = GlobalConfig.instance().getStr("host.public", ConfigGroup.server.getName(), "0.0.0.0");
        String port = GlobalConfig.instance().getStr("port", ConfigGroup.server.getName(), "80");
        List<FileInfo> fileInfos = files.stream().map(o -> {
            FileInfo temp = new FileInfo();
            temp.setUuid(o.getUuid());
            temp.setUrl(StrUtil.format("http://{}:{}/oos/{}", host, port, o.getMd5()));
            temp.setFileName(o.getPath() + "/" + o.getFileName());
            temp.setPublic(o.isPublic());
            return temp;
        }).collect(Collectors.toList());

        ResOosMyFiles resOosMyFiles = new ResOosMyFiles();
        resOosMyFiles.setFileInfos(fileInfos);
        return new IdsResponse<>(resOosMyFiles);
    }


    /**
     * 保存文件到本地
     *
     * @param fileName fileName
     * @param fileData fileData
     */
    private String saveFileToLocal(String fileName, byte[] fileData) throws ErrorCode {
        String workingDir = System.getProperty("user.dir");
        String uploadDir = workingDir + File.separator + "web" + File.separator + "oos";

        // 创建目录（如果不存在）
        FileUtil.mkdir(uploadDir);

        String filePath = uploadDir + File.separator + fileName;
        File file = new File(filePath);
        if (!file.exists()) {
            try (FileOutputStream fos = new FileOutputStream(file)) {
                IoUtil.write(fos, true, fileData);
                log.info("File saved: " + filePath);
            } catch (Exception e) {
                log.error("File upload Exception: " + filePath, e);
                throw ErrorCodeList.FILE_SAVE_FAILED;
            }
        } else {
            log.debug("File already exists, skipping: " + filePath);
        }
        return filePath;
    }

    private void checkToken(String token) throws ErrorCode {
        JWT jwt;
        try {
            jwt = JWTUtil.parseToken(token);
        } catch (Exception e) {
            log.error("parseToken error:", e);
            throw ErrorCodeList.TOKEN_NOT_MATCH;
        }
        JSONObject payloads = jwt.getPayloads();
        String userId = payloads.getStr("userId");
        Long expire = payloads.getLong("expire");
        log.debug("userId:{},expire:{}", userId, expire);
        IdsResponse<ResUserVerify> idsResponse = UserReqTools.verify(userId, new ReqUserVerify(token));
        if (idsResponse == null) {
            throw ErrorCodeList.NETWORK_ERROR;
        }
        ResUserVerify data = idsResponse.getData();
        if (!data.isVerify()) {
            throw ErrorCodeList.TOKEN_NOT_MATCH;
        }
    }
}
